from . import Layer
from ...strategies import Losses
from ...strategies import Metrics


class Cost(Layer):

    def __init__(self, loss, metrics):
        super().__init__()
        self.verbose = 1

        # loss & metrics
        self.f = Losses().select(loss)  # cost function
        self.error = None
        self.metrics = [Metrics().select(metric) for metric in metrics] if metrics else []
        self.score = None

        # parameters
        self.a = None
        self.y = None
        self.dx = None

    def forepropagation(self):
        # check dimension
        if self.y.shape[0] != self.prevlayer.units:
            raise Exception('dimension of y is not fit for units of Output layer')

        # get data
        self.a = self.prevlayer.a

        # loss
        self.error = self.f(self.a, self.y).sum(axis=0, keepdims=True).mean()

        # metrics
        self.score = [metric(self.a, self.y) for metric in self.metrics]

    def backpropagation(self):
        self.dx = self.f(self.a, self.y, derivative=True)



import numpy as np


# loss function factory
class Losses(object):

    def __init__(self):
        self.losses = {
            'mse': mse, 'MSE': mse,

            'binary': binary_cross_entropy,
            'binary_crossentropy': binary_cross_entropy,

            'categorical': categorical_cross_entropy,
            'categorical_crossentropy': categorical_cross_entropy
        }

    def select(self, loss):
        return self.losses[loss]


def mse(a, y, derivative=False):
    if not derivative:
        return 0.5 * (y - a) ** 2
    else:
        return -(y - a)


def binary_cross_entropy(a, y, derivative=False):
    if not derivative:
        return -(y * np.log(a + 1e-08) + (1 - y) * np.log(1 - a + 1e-08))
    else:
        return (a - y) / (a * (1 - a) + 1e-08)


def categorical_cross_entropy(a, y, derivative=False):
    if not derivative:
        return -y * np.log(a + 1e-08)
    else:
        return - y / (a + 1e-08)



import numpy as np


def split(x, y, rate, shuffle=False):
    # shuffle
    data = np.hstack([x, y])
    if shuffle:
        np.random.shuffle(data)
    x, y = np.split(data, [x.shape[1]], axis=1)

    # splitting
    boundary = data.shape[0] - int(np.ceil(data.shape[0] * rate))
    train_x, test_x = np.split(x, [boundary])
    train_y, test_y = np.split(y, [boundary])

    return train_x, train_y, test_x, test_y


def to_class(a, boundary=0.5):
    return np.where(a >= boundary, 1, 0)



import numpy as np


class Activations(object):

    def __init__(self):
        self.activations = {
            'linear': linear,
            'sigmoid': sigmoid,
            'tanh': tanh,
            'relu': relu,
            'leaky_relu': leaky_relu,
            'softmax': softmax
        }

    def select(self, activation):
        return self.activations[activation]


def linear(z, derivative=False):
    if not derivative:
        return z
    else:
        return 1


def sigmoid(z, derivative=False):
    if not derivative:
        return 1 / (1 + np.exp(-z))
    else:
        return sigmoid(z) * (1 - sigmoid(z))


def tanh(z, derivative=False):
    if not derivative:
        return (np.exp(z) - np.exp(-z)) / (np.exp(z) + np.exp(-z))
    else:
        return 1 - tanh(z) ** 2


def relu(x, derivative=False):
    if not derivative:
        return np.max(0, x)
    else:
        return 1 if x > 0 else 0


def leaky_relu(x, derivative=False):
    if not derivative:
        return np.max(0.01*x, x)
    else:
        return 1 if x > 0 else 0.01


def softmax(x, derivative=False):
    if not derivative:
        t = np.exp(x)
        return t / t.sum(0)
    else:
        a = softmax(x)
        n, m = x.shape[0], x.shape[1]
        result = np.empty((m, n, n))

        for index in range(m):
            for x in range(n):
                for y in range(n):
                    if x == y:
                        result[index, x, y] = a[x, index] * (1 - a[x, index])
                    else:
                        result[index, x, y] = -a[x, index] * a[y, index]
        return result



import numpy as np

from .structures.layers import Cost
from .strategies import Optimizers
from .strategies.optimizers import Optimizer
from .utils import split, to_class


class Model(object):

    def __init__(self):
        self.layers = []
        self.optimizer = None

        # logs
        self.losses = {'train': [], 'val': []}
        self.scores = {'train': [], 'val': []}

    # ----------------------------------------------------------------------------------------------------
    # configuration
    # ----------------------------------------------------------------------------------------------------

    def add(self, layer):
        """add layer into model"""

        # add
        self.layers.append(layer)

        # link
        if len(self.layers) > 1:
            self.layers[-1].prevlayer = self.layers[-2]
            self.layers[-2].nextlayer = self.layers[-1]

    def set(self, loss, metrics=None, optimizer='adam'):
        """configure some strategies such as loss, optimizer and metrics"""

        # loss
        if isinstance(loss, str):
            if isinstance(metrics, list) or metrics:  # metrics is list or None
                self.add(Cost(loss, metrics))
            else:
                raise TypeError('metrics should be a list or None')
        else:
            raise TypeError('loss should be a string')

        # optimizer
        if isinstance(optimizer, str):
            self.optimizer = Optimizers().select(optimizer)
        elif isinstance(optimizer, Optimizer):
            self.optimizer = optimizer
        else:
            raise TypeError('optimizer should be given Optimizer object or string')

    # ----------------------------------------------------------------------------------------------------
    # training / testing / predicting
    # ----------------------------------------------------------------------------------------------------

    def train(self, x, y, epochs=3000, size=64, validation=0.0, shuffle=False, verbose=1):
        """the whole processing of training"""

        if shuffle:
            np.random.shuffle(x)
            np.random.shuffle(y)

        # training
        iteration = 0  # counter
        batches = int(np.ceil(x.shape[0] / size))
        for epoch in range(epochs):
            for batch in range(batches):

                # processing data
                start, stop = batch * size, (batch+1) * size
                batch_x, batch_y = x[start:stop], y[start:stop]
                train_x, train_y, val_x, val_y = split(batch_x, batch_y, validation, shuffle=False)

                # train set
                self.__train(train_x, train_y, epoch, batch, iteration=iteration)
                self.__record('train')

                # validation set
                if validation:
                    self.test(val_x, val_y)
                    self.__record('val')

                # report result
                if verbose:
                    result = 'Epoch:{epoch:>4}/{epochs} - Batch:{batch:>2}/{batches}'.format(
                        epoch=epoch+1, epochs=epochs, batch=batch+1, batches=batches
                    )
                    print(result)
                    self.__report('train')
                    if validation:
                        self.__report('val')

                iteration += 1  # total iteration + 1

        if not verbose:  # verbose == 0
            self.__report('train')

    def __train(self, x, y, epoch, batch, **kw):
        """the core processing of training"""

        # feeding data
        self.__feed(x.T, y.T)

        # initialization or updating
        if epoch == batch == 0:
            for layer in self.layers[1:-1]:  # Input & Cost layers have no parameters to initialize
                layer.initialization()
                self.optimizer.initialize(layer)
        else:
            for layer in self.layers[1:-1]:  # Input & Cost layers have no parameters to update
                self.optimizer.update(layer, epoch, **kw)

        # fore propagation
        for layer in self.layers:
            layer.forepropagation()

        # back propagation
        for layer in self.layers[::-1]:
            layer.backpropagation()

    def test(self, x, y):
        """the processing of testing or validation"""

        self.__feed(x.T, y.T)
        for layer in self.layers:
            layer.forepropagation()

        return self.layers[-1].error, self.layers[-1].score

    def predict(self, x, onlyclass=True):
        """the processing of applying trained model to predict something"""

        self.__feed(x.T)
        for layer in self.layers[:-1]:  # no need Cost layer in predicting
            layer.forepropagation()
        y = self.layers[-2].a
        print('prediction:\n', to_class(y) if onlyclass else y)

    def __feed(self, x, y=None):
        """feed data set into layer"""

        self.layers[0].x = x
        self.layers[-1].y = y

    def __record(self, process):
        """record loss and scores in training or validation"""

        self.losses[process].append(self.layers[-1].error)
        self.scores[process].append(self.layers[-1].score)

    def __report(self, process):
        """print results of training or testing"""

        content = 'loss of {process:^5}: {loss:>.7f}   Metrics: {scores}'.format(
            process=process,
            loss=self.losses[process][-1],
            scores=self.scores[process][-1]
        )
        print(content)



from .base import Layer


class Input(Layer):

    def __init__(self, units):
        super().__init__()

        self.units = units

        self.x = None
        self.a = None

    def forepropagation(self):
        # check dimension
        if self.x.shape[0] != self.units:
            raise Exception('dimension of x is not fit for units of Input layer')

        # get data
        self.a = self.x



import numpy as np


class Optimizers(object):

    def __init__(self):
        self.optimizers = {
            'sgd': SGD, 'SGD': SGD,
            'momentum': Momentum,
            'rmsprop': RMSprop, 'RMSprop': RMSprop,
            'adam': Adam, 'Adam': Adam
        }

    def select(self, optimizer):
        return self.optimizers[optimizer]()


class Optimizer(object):

    def __init__(self, alpha=0.01, decay=0.0):
        self.alpha = alpha
        self.decay = decay

    def initialize(self, layer):
        pass

    def update(self, layer, epoch, **kw):
        pass


class SGD(Optimizer):

    def update(self, layer, epoch, **kw):
        alpha = 1 / (1 + self.decay * epoch) * self.alpha
        layer.w -= alpha * layer.dw
        layer.b -= alpha * layer.db


class Momentum(Optimizer):

    def __init__(self, alpha=0.001, decay=0.0, *, beta=0.9):
        super().__init__(alpha, decay)
        self.beta = beta

    def initialize(self, layer):
        layer.vdw = np.zeros_like(layer.w)
        layer.vdb = np.zeros_like(layer.b)

    def update(self, layer, epoch, **kw):
        _alpha = 1 / (1 + self.decay * epoch) * self.alpha
        layer.vdw = self.beta * layer.vdw + (1 - self.beta) * layer.dw
        layer.vdb = self.beta * layer.vdb + (1 - self.beta) * layer.db

        layer.w -= _alpha * layer.vdw
        layer.b -= _alpha * layer.vdb


class RMSprop(Optimizer):

    def __init__(self, alpha=0.001, decay=0.0, *, beta=0.999, epsilon=1e-08):
        super().__init__(alpha, decay)
        self.beta = beta
        self.epsilon = epsilon

    def initialize(self, layer):
        layer.sdw = np.zeros_like(layer.w)
        layer.sdb = np.zeros_like(layer.b)

    def update(self, layer, epoch, **kw):
        _alpha = 1 / (1 + self.decay * epoch) * self.alpha
        layer.sdw = self.beta * layer.sdw + (1 - self.beta) * (layer.dw ** 2)
        layer.sdb = self.beta * layer.sdb + (1 - self.beta) * (layer.db ** 2)

        layer.w -= _alpha * (layer.dw / (np.sqrt(layer.sdw) + self.epsilon))
        layer.b -= _alpha * (layer.db / (np.sqrt(layer.sdb) + self.epsilon))


class Adam(Optimizer):

    def __init__(self, alpha=0.001, decay=0.0, *, beta1=0.9, beta2=0.999, epsilon=1e-08):
        super().__init__(alpha, decay)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon

    def initialize(self, layer):
        layer.vdw = np.zeros_like(layer.w)
        layer.vdb = np.zeros_like(layer.b)
        layer.sdw = np.zeros_like(layer.w)
        layer.sdb = np.zeros_like(layer.b)

    def update(self, layer, epoch, **kw):
        iteration = kw.get('iteration')

        _alpha = 1 / (1 + self.decay * epoch) * self.alpha
        layer.vdw = self.beta1 * layer.vdw + (1 - self.beta1) * layer.dw
        layer.vdb = self.beta1 * layer.vdb + (1 - self.beta1) * layer.db
        layer.sdw = self.beta2 * layer.sdw + (1 - self.beta2) * (layer.dw ** 2)
        layer.sdb = self.beta2 * layer.sdb + (1 - self.beta2) * (layer.db ** 2)

        vcdw = layer.vdw / (1 - self.beta1 ** iteration)
        vcdb = layer.vdb / (1 - self.beta1 ** iteration)
        scdw = layer.sdw / (1 - self.beta2 ** iteration)
        scdb = layer.sdb / (1 - self.beta2 ** iteration)

        layer.w -= _alpha * (vcdw / (np.sqrt(scdw) + self.epsilon))
        layer.b -= _alpha * (vcdb / (np.sqrt(scdb) + self.epsilon))



class Layer(object):

    def __init__(self):
        self.prevlayer = None
        self.nextlayer = None

    def forepropagation(self):
        pass

    def backpropagation(self):
        pass

    def initialization(self):
        pass



import numpy as np
from . import Layer
from ...strategies import Activations


class Dense(Layer):

    def __init__(self, units, activation):
        super().__init__()
        self.units = units
        self.activation = activation
        self.g = Activations().select(activation)

        # fore parameters
        self.x = None
        self.w = None
        self.b = None
        self.z = None
        self.a = None

        # back parameters
        self.da = None
        self.dw = None
        self.db = None
        self.dx = None

    def forepropagation(self):
        # get data
        self.x = self.prevlayer.a

        self.z = self.w @ self.x + self.b
        self.a = self.g(self.z)

    def backpropagation(self):
        # get data
        self.da = self.nextlayer.dx

        dadz = self.g(self.z, derivative=True)
        if self.activation == 'softmax':
            n, m = self.da.shape[0], self.da.shape[1]
            dz = (dadz @ self.da.reshape(m, n, 1)).reshape(m, n).T
        else:
            dz = self.da * dadz

        self.dw = dz @ self.x.T / self.x.shape[1]
        self.db = dz.mean(axis=1, keepdims=True)
        self.dx = self.w.T @ dz

    def initialization(self):
        self.w = np.random.randn(self.units, self.prevlayer.units) * 0.01
        self.b = np.zeros([self.units, 1])



import numpy as np

from ..utils import to_class


class Metrics(object):

    def __init__(self):
        self.metrics = {
            'accuracy': accuracy, 'acc': accuracy,
            'error_rate': error_rate, 'err': error_rate,
            'precision': precision,
            'recall': recall,
            'f1': f1, 'F1': f1
        }

    def select(self, metric):
        return self.metrics[metric]


def accuracy(a, y):
    t = to_class(a)

    right = np.count_nonzero(t == y) / t.shape[0]
    total = t.shape[1]

    return right / total


def error_rate(a, y):
    return 1 - accuracy(a, y)


def precision(a, y):
    t = to_class(a)

    p = np.empty(t.shape[0])
    for i in range(t.shape[0]):
        numerator = np.count_nonzero(y[i][t[i] == 1] == 1)
        denominator = np.count_nonzero(t[i] == 1) + 1e-8
        p[i] = numerator / denominator

    return p


def recall(a, y):
    t = to_class(a)

    r = np.empty(t.shape[0])
    for i in range(t.shape[0]):
        numerator = np.count_nonzero(t[i][y[i] == 1] == 1)
        denominator = np.count_nonzero(y[i] == 1) + 1e-8
        r[i] = numerator / denominator

    return r


def f1(a, y):
    p = precision(a, y)
    r = recall(a, y)
    f = 2 * p * r / (p + r + 1e-08)

    return f
